//=============================================================================================
// Computer Graphics 3rd Homework
//
// A beadott program csak ebben a fajlban lehet, a fajl 1 byte-os ASCII karaktereket tartalmazhat, BOM kihuzando.
// Tilos:
// - mast "beincludolni", illetve mas konyvtarat hasznalni
// - faljmuveleteket vegezni a printf-et kiveve
// - Mashonnan atvett programresszleteket forrasmegjeloles nelkul felhasznalni es
// - felesleges programsorokat a beadott programban hagyni!!!!!!!
// - felesleges kommenteket a beadott programba irni a forrasmegjelolest kommentjeit kiveve
// ---------------------------------------------------------------------------------------------
// A feladatot ANSI C++ nyelvu forditoprogrammal ellenorizzuk, a Visual Studio-hoz kepesti elteresekrol
// es a leggyakoribb hibakrol (pl. ideiglenes objektumot nem lehet referencia tipusnak ertekul adni)
// a hazibeado portal ad egy osszefoglalot.
// ---------------------------------------------------------------------------------------------
// A feladatmegoldasokban csak olyan OpenGL fuggvenyek hasznalhatok, amelyek az oran a feladatkiadasig elhangzottak
// A keretben nem szereplo GLUT fuggvenyek tiltottak.
//
// NYILATKOZAT
// ---------------------------------------------------------------------------------------------
// Nev    : Borcsok Barnabas
// Neptun : WC624D
// ---------------------------------------------------------------------------------------------
// ezennel kijelentem, hogy a feladatot magam keszitettem, es ha barmilyen segitseget igenybe vettem vagy
// mas szellemi termeket felhasznaltam, akkor a forrast es az atvett reszt kommentekben egyertelmuen jeloltem.
// A forrasmegjeloles kotelme vonatkozik az eloadas foliakat es a targy oktatoi, illetve a
// grafhazi doktor tanacsait kiveve barmilyen csatornan (szoban, irasban, Interneten, stb.) erkezo minden egyeb
// informaciora (keplet, program, algoritmus, stb.). Kijelentem, hogy a forrasmegjelolessel atvett reszeket is ertem,
// azok helyessegere matematikai bizonyitast tudok adni. Tisztaban vagyok azzal, hogy az atvett reszek nem szamitanak
// a sajat kontribucioba, igy a feladat elfogadasarol a tobbi resz mennyisege es minosege alapjan szuletik dontes.
// Tudomasul veszem, hogy a forrasmegjeloles kotelmenek megsertese eseten a hazifeladatra adhato pontokat
// negativ elojellel szamoljak el es ezzel parhuzamosan eljaras is indul velem szemben.
//=============================================================================================
#include "framework.h"

//---------------------------
template<class T>
struct Dnum { // Dual numbers for automatic derivation
//---------------------------
    float f; //function value
    T d; //derivatives
    Dnum(float f0 = 0, T d0 = T(0)) { f = f0, d = d0; }

    Dnum operator+(Dnum r) { return Dnum(f + r.f, d + r.d); }

    Dnum operator-(Dnum r) { return Dnum(f - r.f, d - r.d); }

    Dnum operator*(Dnum r) {
        return Dnum(f * r.f, f * r.d + d * r.f);
    }

    Dnum operator/(Dnum r) {
        return Dnum(f / r.f, (r.f * d - r.d * f) / r.f / r.f);
    }
};

template<class T>
Dnum<T> C(float t) { return Dnum<T>(t, 1); }

template<class T>
Dnum<T> Sin(Dnum<T> g) { return Dnum<T>(sinf(g.f), cosf(g.f) * g.d); }

template<class T>
Dnum<T> Cos(Dnum<T> g) { return Dnum<T>(cosf(g.f), -sinf(g.f) * g.d); }

template<class T>
Dnum<T> Sinh(Dnum<T> g) { return Dnum<T>(sinh(g.f), cosh(g.f) * g.d); }

template<class T>
Dnum<T> Cosh(Dnum<T> g) { return Dnum<T>(cosh(g.f), sinh(g.f) * g.d); }

template<class T>
Dnum<T> Tanh(Dnum<T> g) { return Sinh(g) / Cosh(g); }


typedef Dnum<vec2> Dnum2;

// Parametric surface nr of discrete rectangles (triangles/2) during approximation
const int tessellationLevel = 20;

//---------------------------
struct Camera { // 3D camera
//---------------------------
    vec3 wEye, wLookat, wVup;   // extrinsic
    float fov, asp, fp, bp;        // intrinsic
public:
    Camera() {
        asp = (float) windowWidth / windowHeight;
        fov = 75.0f * (float) M_PI / 180.0f;
        fp = 1;
        bp = 42;
    }

    mat4 V() { // view matrix: translates the center to the origin
        vec3 w = normalize(wEye - wLookat);
        vec3 u = normalize(cross(wVup, w));
        vec3 v = cross(w, u);
        return TranslateMatrix(wEye * (-1)) * mat4(u.x, v.x, w.x, 0,
                                                   u.y, v.y, w.y, 0,
                                                   u.z, v.z, w.z, 0,
                                                   0, 0, 0, 1);
    }

    mat4 P() { // projection matrix
        return mat4(1 / (tan(fov / 2) * asp), 0, 0, 0,
                    0, 1 / tan(fov / 2), 0, 0,
                    0, 0, -(fp + bp) / (bp - fp), -1,
                    0, 0, -2 * fp * bp / (bp - fp), 0);
    }

    void Animate(float t) {}
};

//---------------------------
struct Material {
//---------------------------
    vec3 kd, ks, ka;
    float shininess;
};

//---------------------------
struct Light {
//---------------------------
    vec3 La, Le;
    vec4 wLightPos; // homogenous coordinates, can be at ideal point

    void Animate(float t) {}
};

//---------------------------
class CheckerBoardTexture : public Texture {
//---------------------------
public:
    CheckerBoardTexture(const int width = 0, const int height = 0,
                        float c1r = 1, float c1g = 1, float c1b = 0, float c1a = 1,
                        float c2r = 0, float c2g = 0, float c2b = 1, float c2a = 1) : Texture() {
        std::vector<vec4> image(width * height);
        const vec4 c1(c1r, c1g, c1b, c1a), c2(c2r, c2g, c2b, c2a);

        for (int x = 0; x < width; x++)
            for (int y = 0; y < height; y++)
                image[y * width + x] = (x & 1) ^ (y & 1) ? c1 : c2;

        create(width, height, image, GL_NEAREST);
    }
};

//----------------------------
class StripeTexture : public Texture {
//----------------------------
public:
    StripeTexture(const int width = 0, const int height = 0) : Texture() {
        std::vector<vec4> image(width * height);

        const vec4 yellow(1, 1, 0, 1), black(0.1f, 0.1f, 0.1f, 1);

        for (int x = 0; x < width; x++)
            for (int y = 0; y < height; y++)
                image[y * width + x] = x % 3 ? yellow : black;

        create(width, height, image, GL_NEAREST);
    }
};

//---------------------------
struct RenderState {
//---------------------------
    mat4 MVP, M, Minv, V, P;
    Material *material;
    std::vector<Light> lights;
    Texture *texture;
    vec3 wEye;
};

//---------------------------
class Shader : public GPUProgram {
//---------------------------
public:
    virtual void Bind(RenderState state) = 0;

    void setUniformMaterial(const Material &material, const std::string &name) {
        setUniform(material.kd, name + ".kd");
        setUniform(material.ks, name + ".ks");
        setUniform(material.ka, name + ".ka");
        setUniform(material.shininess, name + ".shininess");
    }

    void setUniformLight(const Light &light, const std::string &name) {
        setUniform(light.La, name + ".La");
        setUniform(light.Le, name + ".Le");
        setUniform(light.wLightPos, name + ".wLightPos");
    }
};

//---------------------------
class GouraudShader : public Shader {
//---------------------------
    const char *vertexSource = R"(
		#version 330
		precision highp float;

		struct Light {
			vec3 La, Le;
			vec4 wLightPos;
		};
		
		struct Material {
			vec3 kd, ks, ka;
			float shininess;
		};

		uniform mat4  MVP, M, Minv;  // MVP, Model, Model-inverse
		uniform Light[8] lights;     // light source direction 
		uniform int   nLights;		 // number of light sources
		uniform vec3  wEye;          // pos of eye
		uniform Material  material;  // diffuse, specular, ambient ref

		layout(location = 0) in vec3  vtxPos;            // pos in modeling space
		layout(location = 1) in vec3  vtxNorm;      	 // normal in modeling space

		out vec3 radiance;		    // reflected radiance

		void main() {
			gl_Position = vec4(vtxPos, 1) * MVP; // to NDC
			// radiance computation
			vec4 wPos = vec4(vtxPos, 1) * M;	
			vec3 V = normalize(wEye * wPos.w - wPos.xyz);
			vec3 N = normalize((Minv * vec4(vtxNorm, 0)).xyz);
			if (dot(N, V) < 0) N = -N;	// prepare for one-sided surfaces like Mobius or Klein

			radiance = vec3(0, 0, 0);
			for(int i = 0; i < nLights; i++) {
				vec3 L = normalize(lights[i].wLightPos.xyz * wPos.w - wPos.xyz * lights[i].wLightPos.w);
				vec3 H = normalize(L + V);
				float cost = max(dot(N,L), 0), cosd = max(dot(N,H), 0);
				radiance += material.ka * lights[i].La + (material.kd * cost + material.ks * pow(cosd, material.shininess)) * lights[i].Le;
			}
		}
	)";

    // fragment shader in GLSL
    const char *fragmentSource = R"(
		#version 330
		precision highp float;

		in  vec3 radiance;      // interpolated radiance
		out vec4 fragmentColor; // output goes to frame buffer

		void main() {
			fragmentColor = vec4(radiance, 1);
		}
	)";
public:
    GouraudShader() { create(vertexSource, fragmentSource, "fragmentColor"); }

    void Bind(RenderState state) {
        Use();        // make this program run
        setUniform(state.MVP, "MVP");
        setUniform(state.M, "M");
        setUniform(state.Minv, "Minv");
        setUniform(state.wEye, "wEye");
        setUniformMaterial(*state.material, "material");

        setUniform((int) state.lights.size(), "nLights");
        for (unsigned int i = 0; i < state.lights.size(); i++) {
            setUniformLight(state.lights[i], std::string("lights[") + std::to_string(i) + std::string("]"));
        }
    }
};

//---------------------------
class PhongShader : public Shader {
//---------------------------
    const char *vertexSource = R"(
		#version 330
		precision highp float;

		struct Light {
			vec3 La, Le;
			vec4 wLightPos;
		};

		uniform mat4  MVP, M, Minv; // MVP, Model, Model-inverse
		uniform Light[8] lights;    // light sources 
		uniform int   nLights;
		uniform vec3  wEye;         // pos of eye

		layout(location = 0) in vec3  vtxPos;            // pos in modeling space
		layout(location = 1) in vec3  vtxNorm;      	 // normal in modeling space
		layout(location = 2) in vec2  vtxUV;

		out vec3 wNormal;		    // normal in world space
		out vec3 wView;             // view in world space
		out vec3 wLight[8];		    // light dir in world space
		out vec2 texcoord;

		void main() {
			gl_Position = vec4(vtxPos, 1) * MVP; // to NDC
			// vectors for radiance computation
			vec4 wPos = vec4(vtxPos, 1) * M;
			for(int i = 0; i < nLights; i++) {
				wLight[i] = lights[i].wLightPos.xyz * wPos.w - wPos.xyz * lights[i].wLightPos.w;
			}
		    wView  = wEye * wPos.w - wPos.xyz;
		    wNormal = (Minv * vec4(vtxNorm, 0)).xyz;
		    texcoord = vtxUV;
		}
	)";

    // fragment shader in GLSL
    const char *fragmentSource = R"(
		#version 330
		precision highp float;

		struct Light {
			vec3 La, Le;
			vec4 wLightPos;
		};

		struct Material {
			vec3 kd, ks, ka;
			float shininess;
		};

		uniform Material material;
		uniform Light[8] lights;    // light sources 
		uniform int   nLights;
		uniform sampler2D diffuseTexture;

		in  vec3 wNormal;       // interpolated world sp normal
		in  vec3 wView;         // interpolated world sp view
		in  vec3 wLight[8];     // interpolated world sp illum dir
		in  vec2 texcoord;
		
        out vec4 fragmentColor; // output goes to frame buffer

		void main() {
			vec3 N = normalize(wNormal);
			vec3 V = normalize(wView); 
			if (dot(N, V) < 0) N = -N;	// prepare for one-sided surfaces like Mobius or Klein
			vec3 texColor = texture(diffuseTexture, texcoord).rgb;
			vec3 ka = material.ka * texColor;
			vec3 kd = material.kd * texColor;

			vec3 radiance = vec3(0, 0, 0);
			for(int i = 0; i < nLights; i++) {
				vec3 L = normalize(wLight[i]);
				vec3 H = normalize(L + V);
				float cost = max(dot(N,L), 0), cosd = max(dot(N,H), 0);
				// kd and ka are modulated by the texture
				radiance += ka * lights[i].La + 
                           (kd * texColor * cost + material.ks * pow(cosd, material.shininess)) * lights[i].Le;
			}
			fragmentColor = vec4(radiance, 1);
		}
	)";
public:
    PhongShader() { create(vertexSource, fragmentSource, "fragmentColor"); }

    void Bind(RenderState state) {
        Use();        // make this program run
        setUniform(state.MVP, "MVP");
        setUniform(state.M, "M");
        setUniform(state.Minv, "Minv");
        setUniform(state.wEye, "wEye");
        setUniform(*state.texture, std::string("diffuseTexture"));
        setUniformMaterial(*state.material, "material");

        setUniform((int) state.lights.size(), "nLights");
        for (unsigned int i = 0; i < state.lights.size(); i++) {
            setUniformLight(state.lights[i], std::string("lights[") + std::to_string(i) + std::string("]"));
        }
    }
};

struct VertexData {
    vec3 position, normal;
    vec2 texcoord;
};

//---------------------------
class Geometry {
//---------------------------
protected:
    unsigned int vao, vbo;
public:
    Geometry() {
        glGenVertexArrays(1, &vao);
        glBindVertexArray(vao);
        glGenBuffers(1, &vbo); // Generate 1 vertex buffer object
        glBindBuffer(GL_ARRAY_BUFFER, vbo);
    }

    virtual void Draw() = 0;

    virtual VertexData GenVertexData(float u, float v) {
        throw std::bad_exception();
    };

    ~Geometry() {
        glDeleteBuffers(1, &vbo);
        glDeleteVertexArrays(1, &vao);
    }
};

bool gameOver;

//---------------------------
class ParamSurface : public Geometry {
//---------------------------
    unsigned int nVtxPerStrip, nStrips;

public:
    ParamSurface() { nVtxPerStrip = nStrips = 0; }

    virtual void eval(Dnum2 &U, Dnum2 &V, Dnum2 &X, Dnum2 &Y, Dnum2 &Z) = 0;

    virtual VertexData GenVertexData(float u, float v) {
        VertexData vtxData;
        vtxData.texcoord = vec2(u, v);
        Dnum2 X, Y, Z;
        Dnum2 U(u, vec2(1, 0)), V(v, vec2(0, 1));
        eval(U, V, X, Y, Z);
        vtxData.position = vec3(X.f, Y.f, Z.f);
        vec3 drdU(X.d.x, Y.d.x, Z.d.x), drdV(X.d.y, Y.d.y, Z.d.y);
        vtxData.normal = cross(drdU, drdV);
        return vtxData;
    };

    void create(int N = tessellationLevel, int M = tessellationLevel) {
        nVtxPerStrip = (M + 1) * 2;
        nStrips = N;
        std::vector<VertexData> vtxData;    // vertices on the CPU
        for (int i = 0; i < N; i++) {
            for (int j = 0; j <= M; j++) {
                vtxData.push_back(GenVertexData((float) j / M, (float) i / N));
                vtxData.push_back(GenVertexData((float) j / M, (float) (i + 1) / N));
            }
        }
        glBufferData(GL_ARRAY_BUFFER, nVtxPerStrip * nStrips * sizeof(VertexData), &vtxData[0], GL_STATIC_DRAW);
        // Enable the vertex attribute arrays
        glEnableVertexAttribArray(0);  // attribute array 0 = POSITION
        glEnableVertexAttribArray(1);  // attribute array 1 = NORMAL
        glEnableVertexAttribArray(2);  // attribute array 2 = TEXCOORD0
        // attribute array, components/attribute, component type, normalize?, stride, offset
        glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(VertexData), (void *) offsetof(VertexData, position));
        glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, sizeof(VertexData), (void *) offsetof(VertexData, normal));
        glVertexAttribPointer(2, 2, GL_FLOAT, GL_FALSE, sizeof(VertexData), (void *) offsetof(VertexData, texcoord));
    }

    void Draw() {
        glBindVertexArray(vao);
        for (unsigned int i = 0; i < nStrips; i++) glDrawArrays(GL_TRIANGLE_STRIP, i * nVtxPerStrip, nVtxPerStrip);
    }
};

//---------------------------
class Sphere : public ParamSurface {
//---------------------------
public:
    Sphere() { create(); }

    void eval(Dnum2 &U, Dnum2 &V, Dnum2 &X, Dnum2 &Y, Dnum2 &Z) {
        U = U * 2.0f * (float) M_PI;
        V = V * (float) M_PI;

        X = Cos(U) * Sin(V);
        Y = Sin(U) * Sin(V);
        Z = Cos(V);
    }
};

class VirusBody : public ParamSurface {
public:
    float t;

    VirusBody(float _t = 0) : t(_t) { create(); }

    void eval(Dnum2 &U, Dnum2 &V, Dnum2 &X, Dnum2 &Y, Dnum2 &Z) {
        U = U * 2.0f * (float) M_PI;
        V = V * (float) M_PI;

        X = getRadius(U, V) * Cos(U) * Sin(V);
        Y = getRadius(U, V) * Sin(U) * Sin(V);
        Z = getRadius(U, V) * Cos(V);
    }

    Dnum2 getRadius(Dnum2 U, Dnum2 V) {
        return C<vec2>(1) + C<vec2>(0.06f * (sinf(0.6f * t) + 1)) * Sin(C<vec2>(2) * U + C<vec2>(3) * V + 0.6f * t);
    }
};

//----------------------------
class Tractricoid : public ParamSurface {
//----------------------------
public:
    Tractricoid() { create(); }

    void eval(Dnum2 &U, Dnum2 &V, Dnum2 &X, Dnum2 &Y, Dnum2 &Z) {
        const float height = 3.0f;
        U = U * height;
        V = V * 2 * M_PI;

        X = Cos(V) / Cosh(U);
        Y = Sin(V) / Cosh(U);
        Z = U - Tanh(U);
    }
};

//---------------------------
struct Object {
//---------------------------
    Shader *shader;
    Material *material;
    Texture *texture;
    Geometry *geometry;
    vec3 scale, translation, rotation_axis;
    float rotation_angle;
public:
    Object(Shader *_shader, Material *_material, Texture *_texture, Geometry *_geometry) :
        scale(vec3(1, 1, 1)), translation(vec3(0, 0, 0)), rotation_axis(0, 0, 1), rotation_angle(0) {
        shader = _shader;
        texture = _texture;
        material = _material;
        geometry = _geometry;
    }

    virtual void SetModelingTransform(mat4 &M, mat4 &Minv) {
        M = ScaleMatrix(scale) * RotationMatrix(rotation_angle, rotation_axis) * TranslateMatrix(translation);
        Minv = TranslateMatrix(-translation) * RotationMatrix(-rotation_angle, rotation_axis) *
               ScaleMatrix(vec3(1 / scale.x, 1 / scale.y, 1 / scale.z));
    }

    void Draw(RenderState state) {
        mat4 M, Minv;
        SetModelingTransform(M, Minv);
        state.M = M;
        state.Minv = Minv;
        state.MVP = state.M * state.V * state.P;
        state.material = material;
        state.texture = texture;
        shader->Bind(state);
        geometry->Draw();
    }

    // Animation of the objects
    virtual void Animate(float tstart, float tend) {}
};

struct VirusObject : public Object {

    VirusObject(Shader *_shader, Material *_material, Texture *_texture, VirusBody *_geometry)
        : Object(_shader, _material, _texture, _geometry) {}

    virtual void Animate(float tstart, float tend) {
        if (gameOver) return;
        free(geometry);
        geometry = new VirusBody(tend * 5.0f);
        rotation_angle = 0.8f * tend;
    }
};

struct CoronaObject : public Object {
    Object *parent;
    vec2 uv_pos;

    CoronaObject(Object *_parent, vec2 _uv_pos, Shader *_shader, Material *_material, Texture *_texture,
                 Geometry *_geometry)
        : Object(_shader, _material, _texture, _geometry), parent(_parent), uv_pos(_uv_pos) {}

    void Animate(float tstart, float tend) override {
        VertexData vtxData = parent->geometry->GenVertexData(uv_pos.x, uv_pos.y);
        vtxData.normal = normalize(vtxData.normal);

        vec3 k_hat(0, 0, 1);
        this->rotation_axis = cross(k_hat, vtxData.normal);
        this->rotation_angle = acosf(dot(k_hat, vtxData.normal));
        this->scale = vec3(0.15f, 0.15f, 0.15f);
        this->translation = vtxData.position - vtxData.normal * 0.3f;
    }

    void SetModelingTransform(mat4 &M, mat4 &Minv) override {
        M = ScaleMatrix(scale) *
            RotationMatrix(rotation_angle, rotation_axis) *
            TranslateMatrix(translation) *
            RotationMatrix(parent->rotation_angle, parent->rotation_axis) *
            TranslateMatrix(parent->translation);

        Minv = TranslateMatrix(-parent->translation) *
               RotationMatrix(-parent->rotation_angle, parent->rotation_axis) *
               TranslateMatrix(-translation) *
               RotationMatrix(-rotation_angle, rotation_axis) *
               ScaleMatrix(vec3(1 / scale.x, 1 / scale.y, 1 / scale.z));

    }
};

struct Tetrahedron : public Geometry {
    Tetrahedron() = default;

    //contains the triangles
    std::vector<VertexData> vtx_data;

    vec3 add_triangle_data(vec3 a, vec3 b, vec3 c) {
        auto add_vtx = [this](vec3 p, vec3 n) {
            VertexData curr_vtx;
            curr_vtx.texcoord = vec2(0, 0);
            curr_vtx.position = p;
            curr_vtx.normal = n;
            return curr_vtx;
        };

        vec3 curr_normal = normalize(-cross(a - b, a - c));
        vec3 mid((a + b + c) / 3);
        vec3 origin(0, 0, 0);

        vtx_data.push_back(add_vtx(a, curr_normal));
        vtx_data.push_back(add_vtx(b, curr_normal));
        vtx_data.push_back(add_vtx(c, curr_normal));

        return curr_normal;
    }

    void build_koch_on_triangle(vec3 a, vec3 b, vec3 c, float h, vec3 curr_norm, int depth) {
        vec3 n_a = (a + b) / 2;
        vec3 n_b = (b + c) / 2;
        vec3 n_c = (a + c) / 2;
        vec3 N = (n_a + n_b + n_c) / 3 + curr_norm * h;
        build_koch_at(n_a, n_b, n_c, N, h * 0.65f, depth + 1);
    }

    void build_koch_at(vec3 a, vec3 b, vec3 c, vec3 d, float h, int depth) {
        vec3 n1 = add_triangle_data(a, b, c);
        vec3 n2 = add_triangle_data(a, b, d);
        vec3 n3 = add_triangle_data(a, d, c);
        vec3 n4 = add_triangle_data(b, c, d);

        if (depth > 2) return;

        vec3 N;
        vec3 n_a;
        vec3 n_b;
        vec3 n_c;

        //a b c
        build_koch_on_triangle(a, b, c, h, n1, depth);
        build_koch_on_triangle(a, b, d, h, n2, depth);
        build_koch_on_triangle(a, d, c, h, n3, depth);
        build_koch_on_triangle(b, c, d, h, n4, depth);
    }

    void create(float h) {

        vec3 a(1, 0, -M_SQRT1_2); // 1/sqrt(2) == M_SQRT1_2
        vec3 b(-1, 0, -M_SQRT1_2);
        vec3 c(0, 1, M_SQRT1_2);
        vec3 d(0, -1, M_SQRT1_2);

        vec3 origin{0, 0, 0};

        build_koch_at(a, b, c, d, h, 0);

        glBufferData(GL_ARRAY_BUFFER, vtx_data.size() * sizeof(VertexData), &vtx_data[0], GL_STATIC_DRAW);
        // Enable the vertex attribute arrays
        glEnableVertexAttribArray(0);  // attribute array 0 = POSITION
        glEnableVertexAttribArray(1);  // attribute array 1 = NORMAL
        glEnableVertexAttribArray(2);  // attribute array 2 = TEXCOORD0
        // attribute array, components/attribute, component type, normalize?, stride, offset
        glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(VertexData), (void *) offsetof(VertexData, position));
        glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, sizeof(VertexData), (void *) offsetof(VertexData, normal));
        glVertexAttribPointer(2, 2, GL_FLOAT, GL_FALSE, sizeof(VertexData), (void *) offsetof(VertexData, texcoord));
    }

    void Draw() {
        glBindVertexArray(vao);
        glDrawArrays(GL_TRIANGLES, 0, vtx_data.size());
    }
};

enum AntiBodyDirection {
    x, X, y, Y, z, Z, n
};

struct AntiBodyObject : public Object {

    float dt_brown = 0;
    vec3 curr_dir;
    float v;
    AntiBodyDirection pressed_dir = n;

    AntiBodyObject(Shader *_shader, Material *_material, Texture *_texture,
                   Geometry *_geometry)
        : Object(_shader, _material, _texture, _geometry), curr_dir(0, 0, 0), v(0) {
    }

    void Animate(float tstart, float tend) override {
        free(geometry);
        Tetrahedron *th = new Tetrahedron;
        th->create(fabs(sin(2.0f * tend) - 0.2f) + 0.2f);
        geometry = th;

        float dt = tend - tstart;
        translation = translation + curr_dir * dt * v;

        //move to a random direction every 0.1 sec
        dt_brown += dt;

        if (dt_brown < 0.1f) return;
        //recalculate brown motion
        dt_brown = 0.0f;
        v = (float) (rand() % 6 + 3) / 6;
        int rn = rand() % 12;

        switch (pressed_dir) {
            case x:
                curr_dir =
                    rn > 4 ? vec3( 1, 0, 0) : rn > 3 ? vec3(0, 1, 0) : rn > 2 ? vec3(0, 0, 1) :
                    rn > 1 ? vec3(-1, 0, 0) : rn > 0 ? vec3(0,-1, 0) : vec3(0, 0, -1);
                break;
            case X:
                curr_dir =
                    rn > 4 ? vec3(-1, 0, 0) : rn > 3 ? vec3(0, 1, 0) : rn > 2 ? vec3(0, 0, 1) :
                    rn > 1 ? vec3(1, 0, 0) : rn > 0 ? vec3(0,-1,0): vec3(0,0,-1);
                break;
            case y:
                curr_dir =
                    rn > 4 ? vec3(0, 1, 0) : rn > 3 ? vec3(1, 0, 0) : rn > 2 ? vec3(0, 0, 1) :
                    rn > 1 ? vec3(-1, 0, 0) : rn > 0 ? vec3(0,-1,0): vec3(0,0,-1);
                break;
            case Y:
                curr_dir =
                    rn > 4 ? vec3(0, -1, 0) : rn > 3 ? vec3(0, 1, 0) : rn > 2 ? vec3(0, 0, 1) :
                    rn > 1 ? vec3(-1, 0, 0) : rn > 0 ? vec3(1,0,0): vec3(0,0,-1);
                break;
            case z:
                curr_dir =
                    rn > 4 ? vec3(0, 0, 1) : rn > 3 ? vec3(0, 1, 0) : rn > 2 ? vec3(1, 0, 0) :
                    rn > 1 ? vec3(-1, 0, 0) : rn > 0 ? vec3(0,-1,0): vec3(0,0,-1);
                break;
            case Z:
                curr_dir =
                    rn > 4 ? vec3(0, 0, -1) : rn > 3 ? vec3(0, 1, 0) : rn > 2 ? vec3(0, 0, 1) :
                    rn > 1 ? vec3(-1, 0, 0) : rn > 0 ? vec3(0,-1,0): vec3(1,0,0);
                break;
            case n:
                curr_dir =
                    rn > 9 ? vec3(1, 0, 0) : rn > 7 ? vec3(0, 1, 0) : rn > 5 ? vec3(0, 0, 1) :
                    rn > 3 ? vec3(-1, 0, 0) : rn > 1 ? vec3(0,-1,0): vec3(0,0,-1);
                break;
        }
    }
};

const int nrOfCoronasAtEquator = 20;

//---------------------------
class Scene {
//---------------------------
    std::vector<Object *> objects;
    Camera camera; // 3D camera
    std::vector<Light> lights;
    Object *virusObject;
    AntiBodyObject *anti_body_object;
public:
    void Build() {
        // Shaders
        Shader *phongShader = new PhongShader();
        Shader *gouraudShader = new GouraudShader();

        //-------------------
        // Virus Body
        //-------------------
        Texture *textureVirusBody = new StripeTexture(80, 1);
        Material *materialVirus;
        {
            materialVirus = new Material;
            materialVirus->kd = vec3(0.6f, 0.4f, 0.2f);
            materialVirus->ks = vec3(4, 4, 4);
            materialVirus->ka = vec3(0.1f, 0.1f, 0.1f);
            materialVirus->shininess = 100;
        }

        virusObject = new VirusObject(phongShader, materialVirus, textureVirusBody, new VirusBody());
        virusObject->translation = vec3(-3, 2, 0);
        virusObject->rotation_axis = vec3(0, 1, 1);
        objects.push_back(virusObject);

        //-------------------
        // Virus Coronas
        //-------------------
        Geometry *tractricoid = new Tractricoid();
        Texture *textureCorona = new CheckerBoardTexture(7, 3);
        Material *materialCorona = new Material;
        {
            materialCorona->kd = vec3(0.8f, 0.6f, 0.4f);
            materialCorona->ks = vec3(0.3f, 0.3f, 0.3f);
            materialCorona->ka = vec3(0.2f, 0.2f, 0.2f);
            materialCorona->shininess = 30;
        }

        //generating corona position UV coords
        float v_dist = 1.0f / (float) nrOfCoronasAtEquator;
        for (float v = 0.01f; v < 0.99f; v += v_dist) {
            float curr_dist = 1 / ((float) nrOfCoronasAtEquator * sinf(M_PI * v));
            for (float u = 0.01f; u < 0.99f; u += curr_dist) {
                vec2 curr_uv{u, v};
                Object *corona1 = new CoronaObject(virusObject, curr_uv, phongShader, materialCorona, textureCorona,
                                                   tractricoid);
                corona1->Animate(0, 0);
                objects.push_back(corona1);
            }
        }

        //-------------------
        // Anti Body
        //-------------------
        Tetrahedron *tetrahedron = new Tetrahedron();
        tetrahedron->create(0.8f);
        Texture *texture_anti_body = new CheckerBoardTexture(1, 1);
        Material *material_anti_body = new Material;
        {
            material_anti_body->kd = vec3(0.8f, 0.6f, 0.4f);
            material_anti_body->ks = vec3(0.3f, 0.3f, 0.3f);
            material_anti_body->ka = vec3(0.2f, 0.2f, 0.2f);
            material_anti_body->shininess = 30;
        }

        anti_body_object = new AntiBodyObject(phongShader, material_anti_body, texture_anti_body, tetrahedron);
        objects.push_back(anti_body_object);

        //-------------------
        // Environment
        //-------------------
        Geometry *sphere = new Sphere();
        Texture *textureEnvironment;
        Material *materialEnvironment;
        {
            textureEnvironment = new CheckerBoardTexture(15, 20, 0.1f, 0.1f, 0.4f, 1.0f, 0.1f, 0.1f, 0.5f, 1.0f);
            materialEnvironment = new Material;
            materialEnvironment->kd = vec3(0.2f, 0.2f, 0.4f);
            materialEnvironment->ks = vec3(0.3f, 0.3f, 0.3f);
            materialEnvironment->ka = vec3(0.2f, 0.2f, 0.2f);
            materialEnvironment->shininess = 24;
        }

        Object *environmentSphere = new Object(gouraudShader, materialEnvironment, textureEnvironment, sphere);
        environmentSphere->scale = vec3(10.0f, 10.0f, 10.0f);
        objects.push_back(environmentSphere);

        //-------------------
        // Camera
        //-------------------
        camera.wEye = vec3(0, 0, 8);
        camera.wLookat = vec3(0, 0, 0);
        camera.wVup = vec3(0, 1, 0);

        // Lights
        lights.resize(3);
        lights[0].wLightPos = vec4(5, 5, 4, 0);    // ideal point -> directional light source
        lights[0].La = vec3(0.1f, 0.1f, 1);
        lights[0].Le = vec3(3, 0, 0);

        lights[1].wLightPos = vec4(5, 10, 20, 0);    // ideal point -> directional light source
        lights[1].La = vec3(0.2f, 0.2f, 0.2f);
        lights[1].Le = vec3(0, 3, 0);

        lights[2].wLightPos = vec4(-5, 5, 5, 0);    // ideal point -> directional light source
        lights[2].La = vec3(0.1f, 0.1f, 0.1f);
        lights[2].Le = vec3(0, 0, 3);
    }

    void Render() {
        RenderState state;
        state.wEye = camera.wEye;
        state.V = camera.V();
        state.P = camera.P();
        state.lights = lights;
        for (Object *obj : objects) obj->Draw(state);
    }

    void Animate(float tstart, float tend) {
        camera.Animate(tend);
        for (auto & light : lights) { light.Animate(tend); }
        for (Object *obj : objects) obj->Animate(tstart, tend);
    }

    void CheckForGameOver() {
        if (length(anti_body_object->translation - virusObject->translation) <= 1.5f)
            gameOver = true;
    }

    void handleKeyPress(unsigned char key) {
        switch (key) {
            case 'x':
                anti_body_object->pressed_dir = x;
                break;
            case 'X':
                anti_body_object->pressed_dir = X;
                break;
            case 'y':
                anti_body_object->pressed_dir = y;
                break;
            case 'Y':
                anti_body_object->pressed_dir = Y;
                break;
            case 'z':
                anti_body_object->pressed_dir = z;
                break;
            case 'Z':
                anti_body_object->pressed_dir = Z;
                break;
            default:
                break;
        }
    }

    void handleKeyRelease(unsigned char key) {
        switch (key) {
            case 'x':
                if (anti_body_object->pressed_dir == x) anti_body_object->pressed_dir = n;
                break;
            case 'X':
                if (anti_body_object->pressed_dir == X) anti_body_object->pressed_dir = n;
                break;
            case 'y':
                if (anti_body_object->pressed_dir == y) anti_body_object->pressed_dir = n;
                break;
            case 'Y':
                if (anti_body_object->pressed_dir == Y) anti_body_object->pressed_dir = n;
                break;
            case 'z':
                if (anti_body_object->pressed_dir == z) anti_body_object->pressed_dir = n;
                break;
            case 'Z':
                if (anti_body_object->pressed_dir == Z) anti_body_object->pressed_dir = n;
                break;
            default:
                break;
        }
    }
};

Scene scene;

// Initialization, create an OpenGL context
void onInitialization() {
    glViewport(0, 0, windowWidth, windowHeight);
    glEnable(GL_DEPTH_TEST);
    glDisable(GL_CULL_FACE);
    scene.Build();
}

// Window has become invalid: Redraw
void onDisplay() {
    glClearColor(0.5f, 0.5f, 0.8f, 1.0f);                      // background color
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); // clear the screen
    scene.Render();
    glutSwapBuffers();                                         // exchange the two buffers
}

// Key of ASCII code pressed
void onKeyboard(unsigned char key, int pX, int pY) {
    scene.handleKeyPress(key);
}

// Key of ASCII code released
void onKeyboardUp(unsigned char key, int pX, int pY) {
    scene.handleKeyRelease(key);
}

// Mouse click event
void onMouse(int button, int state, int pX, int pY) {}

// Move mouse with key pressed
void onMouseMotion(int pX, int pY) {
}

// Idle event indicating that some time elapsed: do animation here
void onIdle() {
    static float tend = 0;
    const float dt = 0.1f; // dt is infinitesimal
    float tstart = tend;
    tend = glutGet(GLUT_ELAPSED_TIME) / 1000.0f;

    for (float t = tstart; t < tend; t += dt) {
        float Dt = fmin(dt, tend - t);
        scene.Animate(t, t + Dt);
    }
    scene.CheckForGameOver();
    glutPostRedisplay();
}
